/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/* Elastic pendulum                                                          */
/* ================                                                          */
/*                                                                           */
/* A twodimensional frictionless elastic pendulum consisting of a point      */
/* mass and a spring. The pendulum mass and length, spring constant          */
/* initial angle and other parameters can be changed. The spring is          */
/* ideal and massless.                                                       */
/*                                                                           */
/* Set graphics resolution, console mode or window mode, background colour   */  
/* etc and the pendulum parameters in the file "elpend.cfg".                 */
/*                                                                           */
/* Terminate the program by typing the key "q" (quit).                       */
/*                                                                           */
/* Author etc                                                                */
/* ==========                                                                */
/*                                                                           */
/* Program version: elpend.c 1.00                                            */
/*                                                                           */
/* The program is free to copy and modify.                                   */
/*                                                                           */
/* Author: Stefan Spaennare, Lund, Sweden                                    */
/*                                                                           */
/* E-mail: stefan@spaennare.se                                               */
/*                                                                           */
/* Home page:    http://www.spaennare.se/index.html                          */
/* Program page: http://www.spaennare.se/ssprog2.html                        */
/*                                                                           */
/* C version for Windows: 2006-07-10;   Using Allegro game library and       */
/*                                      Microsoft Visual C++ 6.0.            */
/* Latest update:         2006-07-13                                         */
/*                                                                           */
/*                                                                           */
/* References                                                                */
/* ==========                                                                */
/*                                                                           */
/* 1. The article "The Swinging Spring: A Simple Model of Atmospheric        */
/*    Balance" by Peter Lynch, Met Eireann and Dublin Ireland.               */
/*                                                                           */
/* 2. The program uses a classical 4th order Runge-Kutta method for          */
/*    integration of the differential equations.                             */
/*                                                                           */
/* 3. The program uses a random number generator ran2() from Numerical       */
/*    Recipes in C Second Edition.                                           */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "allegro.h"

double pi=3.14159265358979323;

double timetic=(double)(CLOCKS_PER_SEC);

char cfg_name[21]="elpend.cfg";

char image_name[21]="elpend.bmp";


#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(idum)
long *idum;
{
   int j;
   long k;
   static long idum2=123456789;
   static long iy=0; 
   static long iv[NTAB];
   double temp;

   if (*idum <= 0) {
      if (-(*idum) < 1) {
         *idum=1;
      }
      else {
         *idum=-(*idum);
      } /* if */
      idum2=(*idum);
      for (j=NTAB+7; j>=0; j--) {
         k=(*idum)/IQ1;
	 *idum=IA1*(*idum-k*IQ1)-k*IR1;
	 if (*idum < 0) {
	    *idum += IM1;
	 } /* if */
	 if (j < NTAB) {
	    iv[j]=*idum;
	 } /* if */
      } /* for j */
      iy=iv[0];
   } /* if */
   k=(*idum)/IQ1;
   *idum=IA1*(*idum-k*IQ1)-k*IR1;
   if (*idum < 0) {
      *idum += IM1;
   } /* if */
   k=idum2/IQ2;
   idum2=IA2*(idum2-k*IQ2)-k*IR2;
   if (idum2 < 0) {
      idum2 += IM2;
   } /* if */
   j=iy/NDIV;
   iy=iv[j]-idum2;
   iv[j] = *idum;
   if (iy < 1) {
      iy += IMM1;
   } /* if */
   temp=AM*(double)(iy);
   if (temp > RNMX) {
      return(RNMX);
   }
   else {
      return(temp);
   } /* if */

} /* ran2 */


void comp_der(der1,der2)
double *der1;
double *der2;
{
      double th,pth,r,pr; 
      double l0,k,m,g,temp; 

      th=der1[1];  pth=der1[2];
      r=der1[3];   pr=der1[4];
      l0=der1[5];  k=der1[6];
      m=der1[7];   g=der1[8];

      temp=pth/(m*r*r*r);

      der2[1]=temp*r;
      der2[2]=-m*g*r*sin(th);
      der2[3]=pr/m;
      der2[4]=temp*pth-k*(r-l0)+m*g*cos(th);


} /* comp_der */


int main(argc,argv)
int argc;
char *argv[];
{

   FILE *cfgfile;

   BITMAP *bmp;
   PALETTE pal;

   int i,j,seed,delay0;
   int xsize,ysize,col,c0;
   int wintype,imsave;
   int x0i,y0i,x1i,y1i,x2i,y2i;
   int x1i2,y1i2,rcmi;
   int x1s,y1s,x2s,y2s;
   int col1,col2,bgcol,bordercol,standcol;
   int dispflag1,dispflag2,dispstep;
   int standflag,borderflag,flag;

   long idum;

   double xrf,yrf,xs0,ys0,sum,totsum,ttt1,ttt2;
   double l0,k,m,g,lfac,lf,sfac,standw,standh,rcm;
   double th0,th,pth,r0,r,pr,temp;
   double x1,y1,w,h,inv6,dt;
   double k11,k12,k13,k14,k21,k22,k23,k24;
   double k31,k32,k33,k34,k41,k42,k43,k44;

   double der1[11];
   double der2[11];

   int cr[21];
   int cg[21];
   int cb[21];

   char cc;

   time_t tt;

   clock_t time1,timediff1;


   if ((cfgfile = fopen(cfg_name,"r")) == NULL) {
      printf("File %s not found.\n",cfg_name);
      exit(0);
   } /* if */


   fscanf(cfgfile,"%d %d",&xsize,&ysize);
   fscanf(cfgfile,"%d",&wintype);
   fscanf(cfgfile,"%d",&imsave);

   fscanf(cfgfile,"%d",&bgcol);
   fscanf(cfgfile,"%d",&bordercol);
   fscanf(cfgfile,"%d",&standcol);
   fscanf(cfgfile,"%d",&col1);
   fscanf(cfgfile,"%d",&col2);

   fscanf(cfgfile,"%d",&borderflag);
   fscanf(cfgfile,"%d",&standflag);
   fscanf(cfgfile,"%d",&dispflag1);
   fscanf(cfgfile,"%d",&dispflag2);
   fscanf(cfgfile,"%d",&dispstep);

   fscanf(cfgfile,"%lf",&sfac);
   fscanf(cfgfile,"%lf",&standw);
   fscanf(cfgfile,"%lf",&standh);

   fscanf(cfgfile,"%d",&delay0);
   fscanf(cfgfile,"%lf",&dt);
   fscanf(cfgfile,"%lf",&g);

   fscanf(cfgfile,"%lf",&th0);

   fscanf(cfgfile,"%lf",&lfac);
   fscanf(cfgfile,"%lf",&l0);
   fscanf(cfgfile,"%lf",&r0);

   fscanf(cfgfile,"%lf",&rcm);
   fscanf(cfgfile,"%lf",&m);
   fscanf(cfgfile,"%lf",&k);


   if ((wintype < 0) || (wintype > 1)) {
      printf("wintype must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((imsave < 0) || (imsave > 1)) {
      printf("imsave must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((standflag < 0) || (standflag > 1)) {
      printf("standflag must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((borderflag < 0) || (borderflag > 1)) {
      printf("borderflag must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((dispflag1 < 0) || (dispflag1 > 3)) {
      printf("dispflag1 must be 0, 1, 2 or 3.\n");
      exit(0);
   } /* if */

   if ((dispflag2 < 0) || (dispflag2 > 1)) {
      printf("dispflag2 must be 0 or 1.\n");
      exit(0);
   } /* if */

   if (dispstep < 1) {
      dispstep=1;
   } /* if */


   allegro_init();

   install_keyboard();

   set_color_depth(32);


   if (wintype == 0) {
      if (set_gfx_mode(GFX_AUTODETECT,xsize,ysize,0,0) < 0) {
         if (set_gfx_mode(GFX_AUTODETECT_WINDOWED,xsize,ysize,0,0) < 0) {
            allegro_exit();
            printf("Graphics error or wrong resolution.\n");
            printf("Try to set resolution to 640 480.\n");
            exit(0);
         } /* if */
      } /* if */
   } /* if */

   if (wintype == 1) {
      if (set_gfx_mode(GFX_AUTODETECT_WINDOWED,xsize,ysize,0,0) < 0) {
         if (set_gfx_mode(GFX_AUTODETECT,xsize,ysize,0,0) < 0) {
            allegro_exit();
            printf("Graphics error or wrong resolution.\n");
            printf("Try to set resolution to 640 480.\n");
            exit(0);
         } /* if */
      } /* if */
   } /* if */


   set_palette(desktop_palette);


   cr[0]=0;    cg[0]=0;    cb[0]=0;
   cr[1]=255;  cg[1]=255;  cb[1]=255;

   cr[2]=255;  cg[2]=0;    cb[2]=0;
   cr[3]=0;    cg[3]=255;  cb[3]=0;
   cr[4]=0;    cg[4]=0;    cb[4]=255;

   cr[5]=255;  cg[5]=0;    cb[5]=255;
   cr[6]=255;  cg[6]=255;  cb[6]=0;
   cr[7]=0;    cg[7]=255;  cb[7]=255;

   cr[8]=128;  cg[8]=0;    cb[8]=0;
   cr[9]=0;    cg[9]=128;  cb[9]=0;
   cr[10]=0;   cg[10]=0;   cb[10]=128;

   cr[11]=128; cg[11]=128; cb[11]=0;
   cr[12]=128; cg[12]=0;   cb[12]=128;
   cr[13]=0;   cg[13]=128; cb[13]=128;

   cr[14]=128; cg[14]=128; cb[14]=128;
   cr[15]=160; cg[15]=160; cb[15]=160;
   cr[16]=192; cg[16]=192; cb[16]=192;


   c0=bgcol;
   col=makecol(cr[c0],cg[c0],cb[c0]);

   for (i=0; i<ysize; i++) {
      line(screen,0,i,xsize-1,i,col);
   } /* for i */ 


   tt=time(0);
   ttt1=(double)(tt)+0.5;
   ttt2=fmod(ttt1,2147483647.0);
   seed=(long)(ttt2);
   idum=-seed-1;


   xrf=(double)(xsize)/640.0;
   yrf=(double)(ysize)/480.0;
   xs0=320.0;
   ys0=240.0;

   x0i=(int)(xrf*xs0+0.5);
   y0i=(int)(2.0*(1.0-sfac)*yrf*ys0+0.5);
   x1i=x0i;
   y1i=y0i;
   x2i=x0i;
   y2i=y0i;

   x1s=(int)(2.0*standw*sfac*xrf*xs0+0.5);
   y1s=(int)(2.0*(1.0-sfac)*yrf*ys0-2.0*standh*sfac*yrf*ys0+0.5);
   x2s=-x1s+x0i;
   x1s=x1s+x0i;
   y2s=y1s;

   lf=2.0*lfac*sfac*ys0/l0;

   rcmi=(int)(lf*yrf*rcm+0.5);

   if (th0 < 0.0) {
      temp=fabs(th0);
      th0=(90.0-temp)*ran2(&idum)+temp;
   } /* if */

   th=th0*pi/180.0;
   pth=0.0;
   r=r0;
   pr=0.0;

   der1[5]=l0; der1[6]=k;
   der1[7]=m;  der1[8]=g; 

   h=dt;
   
   inv6=1.0/6.0;

   w=-90.0*pi/180.0;

   totsum=0.0;

   flag=0;

   i=1;


   do {

      
      /* Start of 4th order Runge-Kutta integration */


      der1[1]=th; der1[2]=pth;
      der1[3]=r;  der1[4]=pr;

      comp_der(der1,der2);

      k11=h*der2[1];
      k12=h*der2[2];
      k13=h*der2[3];
      k14=h*der2[4];

      der1[1]=th+0.5*k11; der1[2]=pth+0.5*k12;
      der1[3]=r+0.5*k13;  der1[4]=pr+0.5*k14;

      comp_der(der1,der2);

      k21=h*der2[1];
      k22=h*der2[2];
      k23=h*der2[3];
      k24=h*der2[4];

      der1[1]=th+0.5*k21; der1[2]=pth+0.5*k22;
      der1[3]=r+0.5*k23;  der1[4]=pr+0.5*k24;

      comp_der(der1,der2);

      k31=h*der2[1];
      k32=h*der2[2];
      k33=h*der2[3];
      k34=h*der2[4];

      der1[1]=th+0.5*k31; der1[2]=pth+0.5*k32;
      der1[3]=r+0.5*k33;  der1[4]=pr+0.5*k34;

      comp_der(der1,der2);

      k41=h*der2[1];
      k42=h*der2[2];
      k43=h*der2[3];
      k44=h*der2[4];

      th=th+inv6*(k11+2.0*k21+2.0*k31+k41);
      pth=pth+inv6*(k12+2.0*k22+2.0*k32+k42);
      r=r+inv6*(k13+2.0*k23+2.0*k33+k43);
      pr=pr+inv6*(k14+2.0*k24+2.0*k34+k44);


      /* End of 4th order Runge-Kutta integration. */


      /* Start of display modes. */


      if (i == dispstep) {

         x1=r*cos(th+w);
         y1=r*sin(th+w);

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag2 == 1) {
            circle(screen,x1i,y1i,rcmi,col);
         } /* if */

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag1 == 1) {
            line(screen,x0i,y0i,x1i,y1i,col);
         } /* if */

         x1i=(int)(xrf*lf*x1+0.5)+x0i;
         y1i=-(int)(yrf*lf*y1+0.5)+y0i;

         if (standflag == 1) {
            c0=standcol;
            col=makecol(cr[c0],cg[c0],cb[c0]);
            line(screen,x1s,ysize-1,x1s,y1s,col);
            line(screen,x2s,ysize-1,x2s,y2s,col);
            line(screen,x1s,y1s,x2s,y2s,col);
         } /* if */

         if (flag == 0) {
            x1i2=x1i;
            y1i2=y1i;
         } /* if */

         c0=col1;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag1 == 1) {
            line(screen,x0i,y0i,x1i,y1i,col);
         } /* if */

         if (dispflag1 == 2) {
            putpixel(screen,x1i,y1i,col);
         } /* if */

         if (dispflag1 == 3) {
            line(screen,x1i2,y1i2,x1i,y1i,col);
         } /* if */

         c0=col2;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag2 == 1) {
            circle(screen,x1i,y1i,rcmi,col);
         } /* if */

         if (borderflag == 1) {
            c0=bordercol;
            col=makecol(cr[c0],cg[c0],cb[c0]);
            line(screen,0,0,xsize-1,0,col);
            line(screen,xsize-1,0,xsize-1,ysize-1,col);
            line(screen,xsize-1,ysize-1,0,ysize-1,col);
            line(screen,0,ysize-1,0,0,col);
         } /* if */

         x1i2=x1i;
         y1i2=y1i;

         flag=1;

	 i=0;
	
      } /* if */


      /* End of display modes. */


      i++;

      sum=0.0;

      for (j=1; j<=delay0; j++) {
         sum=sum+1.0/(double)(j);
      } /* for j */

      totsum=totsum+sum;

      if (keypressed()) {
         cc=(readkey() & 0xff);
      } /* if */
      

   } while (cc != 'q');


   if (imsave == 1) {
      get_palette(pal);
      bmp=create_sub_bitmap(screen,0,0,xsize,ysize);
      save_bitmap(image_name,bmp,pal);
      destroy_bitmap(bmp);
   } /* if */

   allegro_exit();

   fclose(cfgfile);

   printf("dummy = %12.6f\n",totsum);

   return(0);

} /* End  */

END_OF_MAIN();
